#include <ctype.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#define MAX_SLOTS 100
#define BUF_SIZE 128

char inputbuf[2*BUF_SIZE];

#ifdef USE_DLMALLOC
#define PROMPT "test-dlmalloc"
#define PWNDBG_COMPAT 1
#include "malloc-2.7.2.c"
#else
#define PROMPT "test-malloc"
#endif

void *slot[MAX_SLOTS];

static int interactive = 0;
static int error = 0;

void dummy()
{
	printf("called dummy\n");
}

void dummy2(const char *p)
{
	printf("called dummy2(\"%s\")\n", p);
}

int findfree()
{
	int i;

	for (i = 0; i < MAX_SLOTS; i++)
		if (!slot[i])
			return i;
	return -1;
}

int checkeol()
{
	if (strtok(NULL, " \t") != NULL) {
		fprintf(stderr, "too many arguments\n");
		error = 1;
		return 0;
	}
	return 1;
}

long getsize()
{
	char *arg;
	long v;

	if ( !(arg = strtok(NULL, " \t")) ) {
		fprintf(stderr, "missing size\n");
		error = 1;
		return -1;
	}
	v = strtol(arg, &arg, 0);
	if (v < 0 || (!isspace(*arg) && *arg != '\0')) {
		fprintf(stderr, "invalid size\n");
		error = 1;
		return -1;
	}
	return v;
}

int getslot()
{
	char *arg;
	long v;

	if ( !(arg = strtok(NULL, " \t")) ) {
		fprintf(stderr, "missing slot index\n");
		error = 1;
		return -1;
	}
	v = strtol(arg, &arg, 0);
	if (v < 0 || v > MAX_SLOTS || (!isspace(*arg) && *arg != '\0')) {
		fprintf(stderr, "invalid slot index\n");
		error = 1;
		return -1;
	}
	return v;
}

int getslotoff(unsigned long *off)
{
	char *arg;
	long v, o;

	if ( !(arg = strtok(NULL, " \t")) ) {
		fprintf(stderr, "missing slot index\n");
		error = 1;
		return -1;
	}
	v = strtol(arg, &arg, 0);
	if (v < 0 || v > MAX_SLOTS || (!isspace(*arg) && *arg != '\0' && *arg != '+')) {
		fprintf(stderr, "invalid slot index\n");
		error = 1;
		return -1;
	}
	if (*arg != '+') {
		*off = 0;
		return v;
	}
	arg++;
	o = strtoul(arg, &arg, 0);
	if (!isspace(*arg) && *arg != '\0') {
		fprintf(stderr, "invalid offset\n");
		error = 1;
		return -1;
	}
	*off = o;
	return v;
}

int puthex(char *dst, char *src)
{
	int n = 0;
	char buf[3];

	while (*src && *(src + 1)) {
		buf[0] = *src++;
		buf[1] = *src++;
		buf[2] = '\0';
		*dst++ = strtol(buf, NULL, 16);
		n++;
	}
	return n;
}

int getcmd(char *buf, int sz)
{
	int c, n;
	char *p;

	if (sz <= 0)
		return 0;

	if (interactive) {
		clearerr(stdin);
		fprintf(stderr, "%s> ", PROMPT);
	}
	n = sz;
	p = buf;
	while (n > 0) {
		c = fgetc(stdin);
		switch (c) {
		case EOF:
			if (n == sz)
				return 0;
			/* fallthrough */
		case '\n':
			*p = '\0';
			return 1;
		default:
			*p++ = c;
			n--;
			break;
		}
	}
	fprintf(stderr, "line too long\n");
	while (!ferror(stdin) && !feof(stdin) && c != '\n')
		c = fgetc(stdin);
	error = 1;
	buf[0] = '\0';
	return interactive;
}

int main()
{
	char buf[BUF_SIZE];
	char *cmd, *arg, *dst;
	size_t l;
	int i ;
	long v, off;

	setvbuf(stdin, inputbuf, _IOLBF, sizeof(inputbuf));
	interactive = isatty(0);

	while ( (!error || interactive) && (l = getcmd(buf, BUF_SIZE)) ) {
		error = 0;
		if ( !(cmd = strtok(buf, " \t")) )
			continue;
		if (!strcmp(cmd, "malloc")) {
			if ( (i = findfree()) < 0 ) {
				fprintf(stderr, "no more empty slots\n");
				error = 1;
				continue;
			}
			if ( (v = getsize()) < 0 )
				continue;
			if (!checkeol(arg))
				continue;
			slot[i] = malloc(v);
			fprintf(stderr, "[%d] %p\n", i, slot[i]);
		} else if (!strcmp(cmd, "free")) {
			if ( (v = getslot()) < 0 )
				continue;
			if (!checkeol(arg))
				continue;
			free(slot[v]);
		} else if (!strcmp(cmd, "list")) {
			if (!checkeol(arg))
				continue;
			for (i = 0; i < MAX_SLOTS; i++) {
				if (slot[i])
					printf("[%d] %p\n", i, slot[i]);
			}
		} else if (!strcmp(cmd, "clear")) {
			if ( (v = getslot()) < 0 )
				continue;
			if (!checkeol(arg))
				continue;
			slot[v] = NULL;
		} else if (!strcmp(cmd, "writestr")) {
			if ( (v = getslotoff(&off)) < 0 )
				continue;
			if ( !(arg = strtok(NULL, " \t")) ) {
				fprintf(stderr, "missing string\n");
				continue;
			}
			if (!checkeol(arg))
				continue;
			strcpy(slot[v] + off, arg);
		} else if (!strcmp(cmd, "writehex")) {
			if ( (v = getslotoff(&off)) < 0 )
				continue;
			dst = slot[v] + off;
			while (arg = strtok(NULL, " \t")) {
				if ( (i = puthex(dst, arg)) < 0 ) {
					fprintf(stderr, "syntax error\n");
					error = 1;
					continue;
				}
				dst += i;
			}
		} else {
			fprintf(stderr, "unknown command\n");
			error = 1;
		}
	}
	printf("\n");
}
